/**
 * Created on 04.01.2005
 */
package spinfo.ql.hmm;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.util.StringTokenizer;

/**
 * To generate the matrix with word-tag transition-probabilities (matrix B)
 * 
 * @author Fabian Steeg
 */
public class MatrixGenerator {
    private HashMap lexicon;

    private HashMap freq;

    private String corpus;

    private double[][] result;

    private String tagset;

    /**
     *  
     */
    public MatrixGenerator(HashMap lexicon, String corpus, String tagset) {
        this.corpus = corpus;
        this.lexicon = lexicon;
        this.tagset = tagset;
        countLexems();
        System.out.println("Tagset: " + tagset);
        System.out.println();
        System.out.println("Corpus: " + corpus);
        System.out.println();
        System.out.println("Frequencies:");
        System.out.println();
        printMap(freq);
    }

    /**
     * @return A matrix computed via Jelinek for word-tag-transitions (Matrix B)
     */
    public double[][] getMatrix() {
        String[] l = new String[lexicon.keySet().size()];
        String[] t = tagset.split(" ");
        int i = 0;
        for (Iterator iter = lexicon.keySet().iterator(); iter.hasNext(); i++) {
            String element = (String) iter.next();
            l[i] = element;
        }
        //init result
        result = new double[l.length][t.length];
        for (int j = 0; j < result.length; j++) {
            Arrays.fill(result[j], 0.0);
            for (int k = 0; k < result[j].length; k++) {
                String categories = ((String) lexicon.get(l[j]));
                if (categories.indexOf(t[k]) != -1) {
                    result[j][k] = 1.0 / categories.split("\\,").length;
                }
            }
        }
//        System.out
//                .println("result after init (lexicon and tags ordered as printed above):");
//        ViterbiMatrixTools.printMatrix(result);
        // adjust:
        for (int j = 0; j < result.length; j++) {
            for (int k = 0; k < result[j].length; k++) {
                double a = result[j][k]
                        * (double) ((Integer) freq.get(l[j])).intValue();
                //compute divider:
                double b = 0.0;
                for (int index = 0; index < result.length; index++) {
                    b = b
                            + result[index][k]
                            * (double) ((Integer) freq.get(l[index]))
                                    .intValue();
                }
                result[j][k] = a / b;
            }
        }
        System.out.println();
        System.out
                .println("Matrix B (word-tag, lexicon and tags ordered as printed above, learned from corpus):");
        ViterbiMatrixTools.printMatrix(result);
        return result;
    }

    
    /**
     * @param diff The variability for the random numbers
     * @return A random-filled matrix for tag-tag-transitions (Matrix A)
     */
    public double[][] getRandomMatrix(double diff) {
        //        String[] l = new String[lexicon.keySet().size()];
        String[] t = tagset.split(" ");
        double[][] res = new double[t.length][t.length];
        int cols = res[0].length;
        double average = 1.0 / (double) cols;
        for (int i = 0; i < res.length; i++) {
            Arrays.fill(res[i], average);
        }
//        System.out.println("random matrix after init:");
//        ViterbiMatrixTools.printMatrix(res);
        Random r = new Random();
        for (int i = 0; i < res.length; i++) {
            double var = 1.0;
            while (var > diff)
                var = r.nextDouble();
            int pos = r.nextInt(res[i].length);
            boolean plus = true;
//            System.out.println("gonna use: " + var);
            for (int j = 0; j < res[i].length; j++) {
                //if not even skip one random position
                if (res[i].length % 2 != 0 && j == pos)
                    continue;
                if (plus && res[i][j] - var > 0) {
//                    System.out.println("changing " + res[i][j] + " to " + ( res[i][j] + var));
                    res[i][j] = res[i][j] + var;
                    plus = false;
                } else if (res[i][j] - var > 0){
//                    System.out.println("changing " + res[i][j] + " to " + ( res[i][j] - var));
                    res[i][j] = res[i][j] - var;
                    plus = true;
                }
            }
        }
        System.out.println();
        System.out.println("Matrix A (tag-tag, pseudo-random-generated):");
        ViterbiMatrixTools.printMatrix(res);
        System.out.println();
        return res;
    }

    /**
     *  
     */
    private void printMap(HashMap map) {
        for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
            String element = (String) iter.next();
            System.out.println(element + ": " + map.get(element));
        }
    }

    /**
     * @param corpus
     */
    private void countLexems() {
        freq = new HashMap();
        StringTokenizer tok = new StringTokenizer(corpus, " \t\n.,", true);
        while (tok.hasMoreTokens()) {
            String rec = tok.nextToken();
            if (lexicon.containsKey(rec)) {
                if (freq.containsKey(rec)) {
                    freq.put(rec, new Integer(((Integer) freq.get(rec))
                            .intValue() + 1));
                } else {
                    freq.put(rec, new Integer(1));
                }
            }
        }
    }
}